let Genode = env:DHALL_GENODE

let Prelude = Genode.Prelude

let Init = Genode.Init

let Child = Init.Child

let TextMapType = Prelude.Map.Type Text

let ChildMapType = TextMapType Child.Type

let Manifest/Type = TextMapType (TextMapType Text)

in  λ ( params
      : { binaries :
            { cached_fs_rom : Text, jitter_sponge : Text, rtc_drv : Text }
        , extraCoreChildren : ChildMapType
        , subinit : Init.Type
        , storeSize : Natural
        , routes : List Init.ServiceRoute.Type
        , bootManifest : Manifest/Type
        }
      ) →
      Genode.Boot::{
      , config = Init::{
        , routes = params.routes
        , children =
            let child = Prelude.Map.keyValue Child.Type

            in    [ child
                      "timer"
                      ( Child.flat
                          Child.Attributes::{
                          , binary = "timer_drv"
                          , config = Init.Config::{
                            , policies =
                              [ Init.Config.Policy::{
                                , service = "Timer"
                                , label = Init.LabelSelector.none
                                }
                              ]
                            }
                          }
                      )
                  , child
                      "rtc"
                      ( Child.flat
                          Child.Attributes::{
                          , binary = params.binaries.rtc_drv
                          , routes = [ Init.ServiceRoute.parent "IO_PORT" ]
                          , config = Init.Config::{
                            , policies =
                              [ Init.Config.Policy::{
                                , service = "Rtc"
                                , label = Init.LabelSelector.none
                                }
                              ]
                            }
                          }
                      )
                  , child
                      "jitter_sponge"
                      ( Child.flat
                          Child.Attributes::{
                          , binary = params.binaries.jitter_sponge
                          , config = Init.Config::{
                            , policies =
                              [ Init.Config.Policy::{
                                , service = "Terminal"
                                , label = Init.LabelSelector.suffix "entropy"
                                }
                              ]
                            }
                          }
                      )
                  , child
                      "store_rom"
                      ( Child.flat
                          Child.Attributes::{
                          , binary = params.binaries.cached_fs_rom
                          , resources = Init.Resources::{
                            , ram = params.storeSize + Genode.units.MiB 1
                            }
                          , config = Init.Config::{
                            , defaultPolicy = Some Init.Config.DefaultPolicy::{
                              , attributes = toMap { directory = "/eris" }
                              }
                            , policies =
                                let directStoreROMs =
                                      Prelude.List.map
                                        Text
                                        Init.Config.Policy.Type
                                        ( λ(childName : Text) →
                                            Init.Config.Policy::{
                                            , service = "ROM"
                                            , label =
                                                Init.LabelSelector.Type.Partial
                                                  { prefix = Some
                                                      "nixos -> ${childName} -> /nix/store/"
                                                  , suffix = None Text
                                                  }
                                            , attributes = toMap
                                                { directory = "/" }
                                            }
                                        )
                                        ( Prelude.Map.keys
                                            Text
                                            Init.Child.Type
                                            params.subinit.children
                                        )

                                in    [ Init.Config.Policy::{
                                        , service = "ROM"
                                        , label =
                                            Init.LabelSelector.prefix
                                              "nixos -> /nix/store"
                                        }
                                      , Init.Config.Policy::{
                                        , attributes = toMap
                                            { directory = "/eris" }
                                        , label =
                                            Init.LabelSelector.prefix
                                              "nixos -> "
                                        , service = "ROM"
                                        }
                                      ]
                                    # directStoreROMs
                            }
                          }
                      )
                  ]
                # params.extraCoreChildren
                # [ child
                      "nixos"
                      ( Init.toChild
                          params.subinit
                          Init.Attributes::{
                          , exitPropagate = True
                          , resources = Init.Resources::{
                            , ram = Genode.units.MiB 4
                            }
                          , routes =
                              let parentROMs =
                                    Prelude.List.concatMap
                                      Text
                                      Init.ServiceRoute.Type
                                      ( λ(suffix : Text) →
                                          Prelude.List.map
                                            Text
                                            Init.ServiceRoute.Type
                                            ( λ(prefix : Text) →
                                                { service =
                                                  { name = "ROM"
                                                  , label =
                                                      Init.LabelSelector.Type.Partial
                                                        { prefix = Some prefix
                                                        , suffix = Some suffix
                                                        }
                                                  }
                                                , route =
                                                    Init.Route.parent
                                                      (Some suffix)
                                                }
                                            )
                                            ( Prelude.Map.keys
                                                Text
                                                Init.Child.Type
                                                params.subinit.children
                                            )
                                      )

                              in    parentROMs
                                      [ "ld.lib.so", "vfs.lib.so", "init" ]
                                  # [ Init.ServiceRoute.parent "IO_MEM"
                                    , Init.ServiceRoute.parent "IO_PORT"
                                    , Init.ServiceRoute.parent "IRQ"
                                    , Init.ServiceRoute.parent "VM"
                                    , Init.ServiceRoute.child "Timer" "timer"
                                    , Init.ServiceRoute.child "Rtc" "rtc"
                                    ]
                          }
                      )
                  ]
        }
      , rom =
          Genode.BootModules.toRomPaths
            ( Prelude.List.concat
                (Prelude.Map.Entry Text Text)
                ( Prelude.Map.values
                    Text
                    (Prelude.Map.Type Text Text)
                    params.bootManifest
                )
            )
      }
